@@ -289,19 +289,11 @@ OutgoingMessage.prototype._send = function _send(data, encoding, callback) {
       data = this._header + data;
     } else {
       const header = this._header;
-      if (this.outputData.length === 0) {
-        this.outputData = [{
-          data: header,
-          encoding: 'latin1',
-          callback: null
-        }];
-      } else {
-        this.outputData.unshift({
-          data: header,
-          encoding: 'latin1',
-          callback: null
-        });
-      }
+      this.outputData.unshift({
+        data: header,
+        encoding: 'latin1',
+        callback: null
+      });
       this.outputSize += header.length;
       this._onPendingData(header.length);
     }
@@ -727,6 +719,7 @@ OutgoingMessage.prototype.addTrailers = function addTrailers(headers) {
 };
 
 function onFinish(outmsg) {
+  if (outmsg && outmsg.socket && outmsg.socket._hadError) return;
   outmsg.emit('finish');
 }
 
