@@ -1,21 +1,22 @@
 'use strict';
-const assert = require('assert');
-const dgram = require('dgram');
+const { Map } = primordials;
+const assert = require('internal/assert');
+const dgram = require('internal/dgram');
 const net = require('net');
 
 module.exports = SharedHandle;
 
-function SharedHandle(key, address, port, addressType, fd, flags) {
+function SharedHandle(key, address, { port, addressType, fd, flags }) {
   this.key = key;
-  this.workers = [];
+  this.workers = new Map();
   this.handle = null;
   this.errno = 0;
 
-  var rval;
+  let rval;
   if (addressType === 'udp4' || addressType === 'udp6')
     rval = dgram._createSocketHandle(address, port, addressType, fd, flags);
   else
-    rval = net._createServerHandle(address, port, addressType, fd);
+    rval = net._createServerHandle(address, port, addressType, fd, flags);
 
   if (typeof rval === 'number')
     this.errno = rval;
@@ -24,20 +25,18 @@ function SharedHandle(key, address, port, addressType, fd, flags) {
 }
 
 SharedHandle.prototype.add = function(worker, send) {
-  assert(this.workers.indexOf(worker) === -1);
-  this.workers.push(worker);
+  assert(!this.workers.has(worker.id));
+  this.workers.set(worker.id, worker);
   send(this.errno, null, this.handle);
 };
 
 SharedHandle.prototype.remove = function(worker) {
-  const index = this.workers.indexOf(worker);
+  if (!this.workers.has(worker.id))
+    return false;
 
-  if (index === -1)
-    return false; // The worker wasn't sharing this handle.
+  this.workers.delete(worker.id);
 
-  this.workers.splice(index, 1);
-
-  if (this.workers.length !== 0)
+  if (this.workers.size !== 0)
     return false;
 
   this.handle.close();
