@@ -1,9 +1,32 @@
 'use strict';
 
-const binding = process.binding('http2');
-const errors = require('internal/errors');
+const {
+  ArrayIsArray,
+  Error,
+  MathMax,
+  Number,
+  ObjectCreate,
+  ObjectKeys,
+  Set,
+  Symbol,
+} = primordials;
+
+const binding = internalBinding('http2');
+const {
+  codes: {
+    ERR_HTTP2_HEADER_SINGLE_VALUE,
+    ERR_HTTP2_INVALID_CONNECTION_HEADERS,
+    ERR_HTTP2_INVALID_PSEUDOHEADER,
+    ERR_HTTP2_INVALID_SETTING_VALUE,
+    ERR_INVALID_ARG_TYPE
+  },
+  addCodeToName,
+  hideStackFrames
+} = require('internal/errors');
 
 const kSocket = Symbol('socket');
+const kProxySocket = Symbol('proxySocket');
+const kRequest = Symbol('request');
 
 const {
   NGHTTP2_SESSION_CLIENT,
@@ -14,6 +37,7 @@ const {
   HTTP2_HEADER_AUTHORITY,
   HTTP2_HEADER_SCHEME,
   HTTP2_HEADER_PATH,
+  HTTP2_HEADER_PROTOCOL,
   HTTP2_HEADER_ACCESS_CONTROL_ALLOW_CREDENTIALS,
   HTTP2_HEADER_ACCESS_CONTROL_MAX_AGE,
   HTTP2_HEADER_ACCESS_CONTROL_REQUEST_METHOD,
@@ -72,7 +96,8 @@ const kValidPseudoHeaders = new Set([
   HTTP2_HEADER_METHOD,
   HTTP2_HEADER_AUTHORITY,
   HTTP2_HEADER_SCHEME,
-  HTTP2_HEADER_PATH
+  HTTP2_HEADER_PATH,
+  HTTP2_HEADER_PROTOCOL
 ]);
 
 // This set contains headers that are permitted to have only a single
@@ -83,6 +108,7 @@ const kSingleValueHeaders = new Set([
   HTTP2_HEADER_AUTHORITY,
   HTTP2_HEADER_SCHEME,
   HTTP2_HEADER_PATH,
+  HTTP2_HEADER_PROTOCOL,
   HTTP2_HEADER_ACCESS_CONTROL_ALLOW_CREDENTIALS,
   HTTP2_HEADER_ACCESS_CONTROL_MAX_AGE,
   HTTP2_HEADER_ACCESS_CONTROL_REQUEST_METHOD,
@@ -149,7 +175,8 @@ const IDX_SETTINGS_INITIAL_WINDOW_SIZE = 2;
 const IDX_SETTINGS_MAX_FRAME_SIZE = 3;
 const IDX_SETTINGS_MAX_CONCURRENT_STREAMS = 4;
 const IDX_SETTINGS_MAX_HEADER_LIST_SIZE = 5;
-const IDX_SETTINGS_FLAGS = 6;
+const IDX_SETTINGS_ENABLE_CONNECT_PROTOCOL = 6;
+const IDX_SETTINGS_FLAGS = 7;
 
 const IDX_SESSION_STATE_EFFECTIVE_LOCAL_WINDOW_SIZE = 0;
 const IDX_SESSION_STATE_EFFECTIVE_RECV_DATA_LENGTH = 1;
@@ -174,10 +201,13 @@ const IDX_OPTIONS_PEER_MAX_CONCURRENT_STREAMS = 3;
 const IDX_OPTIONS_PADDING_STRATEGY = 4;
 const IDX_OPTIONS_MAX_HEADER_LIST_PAIRS = 5;
 const IDX_OPTIONS_MAX_OUTSTANDING_PINGS = 6;
-const IDX_OPTIONS_FLAGS = 7;
+const IDX_OPTIONS_MAX_OUTSTANDING_SETTINGS = 7;
+const IDX_OPTIONS_MAX_SESSION_MEMORY = 8;
+const IDX_OPTIONS_MAX_SETTINGS = 9;
+const IDX_OPTIONS_FLAGS = 10;
 
 function updateOptionsBuffer(options) {
-  var flags = 0;
+  let flags = 0;
   if (typeof options.maxDeflateDynamicTableSize === 'number') {
     flags |= (1 << IDX_OPTIONS_MAX_DEFLATE_DYNAMIC_TABLE_SIZE);
     optionsBuffer[IDX_OPTIONS_MAX_DEFLATE_DYNAMIC_TABLE_SIZE] =
@@ -213,13 +243,28 @@ function updateOptionsBuffer(options) {
     optionsBuffer[IDX_OPTIONS_MAX_OUTSTANDING_PINGS] =
       options.maxOutstandingPings;
   }
+  if (typeof options.maxOutstandingSettings === 'number') {
+    flags |= (1 << IDX_OPTIONS_MAX_OUTSTANDING_SETTINGS);
+    optionsBuffer[IDX_OPTIONS_MAX_OUTSTANDING_SETTINGS] =
+      MathMax(1, options.maxOutstandingSettings);
+  }
+  if (typeof options.maxSessionMemory === 'number') {
+    flags |= (1 << IDX_OPTIONS_MAX_SESSION_MEMORY);
+    optionsBuffer[IDX_OPTIONS_MAX_SESSION_MEMORY] =
+      MathMax(1, options.maxSessionMemory);
+  }
+  if (typeof options.maxSettings === 'number') {
+    flags |= (1 << IDX_OPTIONS_MAX_SETTINGS);
+    optionsBuffer[IDX_OPTIONS_MAX_SETTINGS] =
+      MathMax(1, options.maxSettings);
+  }
   optionsBuffer[IDX_OPTIONS_FLAGS] = flags;
 }
 
 function getDefaultSettings() {
   settingsBuffer[IDX_SETTINGS_FLAGS] = 0;
   binding.refreshDefaultSettings();
-  const holder = Object.create(null);
+  const holder = ObjectCreate(null);
 
   const flags = settingsBuffer[IDX_SETTINGS_FLAGS];
 
@@ -259,10 +304,16 @@ function getDefaultSettings() {
       settingsBuffer[IDX_SETTINGS_MAX_HEADER_LIST_SIZE];
   }
 
+  if ((flags & (1 << IDX_SETTINGS_ENABLE_CONNECT_PROTOCOL)) ===
+      (1 << IDX_SETTINGS_ENABLE_CONNECT_PROTOCOL)) {
+    holder.enableConnectProtocol =
+      settingsBuffer[IDX_SETTINGS_ENABLE_CONNECT_PROTOCOL] === 1;
+  }
+
   return holder;
 }
 
-// remote is a boolean. true to fetch remote settings, false to fetch local.
+// Remote is a boolean. true to fetch remote settings, false to fetch local.
 // this is only called internally
 function getSettings(session, remote) {
   if (remote)
@@ -276,12 +327,14 @@ function getSettings(session, remote) {
     initialWindowSize: settingsBuffer[IDX_SETTINGS_INITIAL_WINDOW_SIZE],
     maxFrameSize: settingsBuffer[IDX_SETTINGS_MAX_FRAME_SIZE],
     maxConcurrentStreams: settingsBuffer[IDX_SETTINGS_MAX_CONCURRENT_STREAMS],
-    maxHeaderListSize: settingsBuffer[IDX_SETTINGS_MAX_HEADER_LIST_SIZE]
+    maxHeaderListSize: settingsBuffer[IDX_SETTINGS_MAX_HEADER_LIST_SIZE],
+    enableConnectProtocol:
+      !!settingsBuffer[IDX_SETTINGS_ENABLE_CONNECT_PROTOCOL]
   };
 }
 
 function updateSettingsBuffer(settings) {
-  var flags = 0;
+  let flags = 0;
   if (typeof settings.headerTableSize === 'number') {
     flags |= (1 << IDX_SETTINGS_HEADER_TABLE_SIZE);
     settingsBuffer[IDX_SETTINGS_HEADER_TABLE_SIZE] =
@@ -311,6 +364,11 @@ function updateSettingsBuffer(settings) {
     flags |= (1 << IDX_SETTINGS_ENABLE_PUSH);
     settingsBuffer[IDX_SETTINGS_ENABLE_PUSH] = Number(settings.enablePush);
   }
+  if (typeof settings.enableConnectProtocol === 'boolean') {
+    flags |= (1 << IDX_SETTINGS_ENABLE_CONNECT_PROTOCOL);
+    settingsBuffer[IDX_SETTINGS_ENABLE_CONNECT_PROTOCOL] =
+      Number(settings.enableConnectProtocol);
+  }
 
   settingsBuffer[IDX_SETTINGS_FLAGS] = flags;
 }
@@ -368,42 +426,42 @@ function isIllegalConnectionSpecificHeader(name, value) {
   }
 }
 
-function assertValidPseudoHeader(key) {
+const assertValidPseudoHeader = hideStackFrames((key) => {
   if (!kValidPseudoHeaders.has(key)) {
-    const err = new errors.Error('ERR_HTTP2_INVALID_PSEUDOHEADER', key);
-    Error.captureStackTrace(err, assertValidPseudoHeader);
-    return err;
+    throw new ERR_HTTP2_INVALID_PSEUDOHEADER(key);
   }
-}
+});
 
-function assertValidPseudoHeaderResponse(key) {
+const assertValidPseudoHeaderResponse = hideStackFrames((key) => {
   if (key !== ':status') {
-    const err = new errors.Error('ERR_HTTP2_INVALID_PSEUDOHEADER', key);
-    Error.captureStackTrace(err, assertValidPseudoHeaderResponse);
-    return err;
+    throw new ERR_HTTP2_INVALID_PSEUDOHEADER(key);
   }
-}
+});
 
-function assertValidPseudoHeaderTrailer(key) {
-  const err = new errors.Error('ERR_HTTP2_INVALID_PSEUDOHEADER', key);
-  Error.captureStackTrace(err, assertValidPseudoHeaderTrailer);
-  return err;
-}
+const assertValidPseudoHeaderTrailer = hideStackFrames((key) => {
+  throw new ERR_HTTP2_INVALID_PSEUDOHEADER(key);
+});
 
 function mapToHeaders(map,
                       assertValuePseudoHeader = assertValidPseudoHeader) {
   let ret = '';
   let count = 0;
-  const keys = Object.keys(map);
+  const keys = ObjectKeys(map);
   const singles = new Set();
-  for (var i = 0; i < keys.length; i++) {
-    let key = keys[i];
-    let value = map[key];
+  let i, j;
+  let isArray;
+  let key;
+  let value;
+  let isSingleValueHeader;
+  let err;
+  for (i = 0; i < keys.length; ++i) {
+    key = keys[i];
+    value = map[key];
     if (value === undefined || key === '')
       continue;
     key = key.toLowerCase();
-    const isSingleValueHeader = kSingleValueHeaders.has(key);
-    let isArray = Array.isArray(value);
+    isSingleValueHeader = kSingleValueHeaders.has(key);
+    isArray = ArrayIsArray(value);
     if (isArray) {
       switch (value.length) {
         case 0:
@@ -414,37 +472,37 @@ function mapToHeaders(map,
           break;
         default:
           if (isSingleValueHeader)
-            return new errors.Error('ERR_HTTP2_HEADER_SINGLE_VALUE', key);
+            throw new ERR_HTTP2_HEADER_SINGLE_VALUE(key);
       }
     } else {
       value = String(value);
     }
     if (isSingleValueHeader) {
       if (singles.has(key))
-        return new errors.Error('ERR_HTTP2_HEADER_SINGLE_VALUE', key);
+        throw new ERR_HTTP2_HEADER_SINGLE_VALUE(key);
       singles.add(key);
     }
     if (key[0] === ':') {
-      const err = assertValuePseudoHeader(key);
+      err = assertValuePseudoHeader(key);
       if (err !== undefined)
-        return err;
+        throw err;
       ret = `${key}\0${value}\0${ret}`;
       count++;
-    } else {
-      if (isIllegalConnectionSpecificHeader(key, value)) {
-        return new errors.Error('ERR_HTTP2_INVALID_CONNECTION_HEADERS', key);
-      }
-      if (isArray) {
-        for (var k = 0; k < value.length; k++) {
-          const val = String(value[k]);
-          ret += `${key}\0${val}\0`;
-        }
-        count += value.length;
-      } else {
-        ret += `${key}\0${value}\0`;
-        count++;
-      }
+      continue;
     }
+    if (isIllegalConnectionSpecificHeader(key, value)) {
+      throw new ERR_HTTP2_INVALID_CONNECTION_HEADERS(key);
+    }
+    if (isArray) {
+      for (j = 0; j < value.length; ++j) {
+        const val = String(value[j]);
+        ret += `${key}\0${val}\0`;
+      }
+      count += value.length;
+      continue;
+    }
+    ret += `${key}\0${value}\0`;
+    count++;
   }
 
   return [ret, count];
@@ -454,45 +512,44 @@ class NghttpError extends Error {
   constructor(ret) {
     super(binding.nghttp2ErrorString(ret));
     this.code = 'ERR_HTTP2_ERROR';
-    this.name = 'Error [ERR_HTTP2_ERROR]';
     this.errno = ret;
+    addCodeToName(this, super.name, 'ERR_HTTP2_ERROR');
+  }
+
+  toString() {
+    return `${this.name} [${this.code}]: ${this.message}`;
   }
 }
 
-function assertIsObject(value, name, types = 'object') {
+const assertIsObject = hideStackFrames((value, name, types) => {
   if (value !== undefined &&
       (value === null ||
        typeof value !== 'object' ||
-       Array.isArray(value))) {
-    const err = new errors.TypeError('ERR_INVALID_ARG_TYPE', name, types);
-    Error.captureStackTrace(err, assertIsObject);
-    throw err;
+       ArrayIsArray(value))) {
+    throw new ERR_INVALID_ARG_TYPE(name, types || 'Object', value);
   }
-}
+});
 
-function assertWithinRange(name, value, min = 0, max = Infinity) {
-  if (value !== undefined &&
+const assertWithinRange = hideStackFrames(
+  (name, value, min = 0, max = Infinity) => {
+    if (value !== undefined &&
       (typeof value !== 'number' || value < min || value > max)) {
-    const err = new errors.RangeError('ERR_HTTP2_INVALID_SETTING_VALUE',
-                                      name, value);
-    err.min = min;
-    err.max = max;
-    err.actual = value;
-    Error.captureStackTrace(err, assertWithinRange);
-    throw err;
+      throw new ERR_HTTP2_INVALID_SETTING_VALUE.RangeError(
+        name, value, min, max);
+    }
   }
-}
+);
 
 function toHeaderObject(headers) {
-  const obj = Object.create(null);
+  const obj = ObjectCreate(null);
   for (var n = 0; n < headers.length; n = n + 2) {
-    var name = headers[n];
-    var value = headers[n + 1];
+    const name = headers[n];
+    let value = headers[n + 1];
     if (name === HTTP2_HEADER_STATUS)
       value |= 0;
-    var existing = obj[name];
+    const existing = obj[name];
     if (existing === undefined) {
-      obj[name] = value;
+      obj[name] = name === HTTP2_HEADER_SET_COOKIE ? [value] : value;
     } else if (!kSingleValueHeaders.has(name)) {
       switch (name) {
         case HTTP2_HEADER_COOKIE:
@@ -511,10 +568,7 @@ function toHeaderObject(headers) {
           // fields with the same name.  Since it cannot be combined into a
           // single field-value, recipients ought to handle "Set-Cookie" as a
           // special case while processing header fields."
-          if (Array.isArray(existing))
-            existing.push(value);
-          else
-            obj[name] = [existing, value];
+          existing.push(value);
           break;
         default:
           // https://tools.ietf.org/html/rfc7230#section-3.2.2
@@ -556,6 +610,8 @@ module.exports = {
   getStreamState,
   isPayloadMeaningless,
   kSocket,
+  kProxySocket,
+  kRequest,
   mapToHeaders,
   NghttpError,
   sessionName,
