@@ -35,32 +35,30 @@ deepStrictEqual(
 );
 
 [
-  [{}, 'object'],
-  [new Boolean(true), 'boolean'],
-  [{ valueOf() { return null; } }, 'object'],
-  [{ valueOf() { return undefined; } }, 'object'],
-  [{ valueOf: null }, 'object'],
-  [Object.create(null), 'object']
-].forEach(([input, actualType]) => {
-  const err = common.expectsError({
+  {},
+  new Boolean(true),
+  { valueOf() { return null; } },
+  { valueOf() { return undefined; } },
+  { valueOf: null },
+  Object.create(null),
+  new Number(true),
+  new MyBadPrimitive(),
+  Symbol(),
+  5n,
+  (one, two, three) => {},
+  undefined,
+  null
+].forEach((input) => {
+  const errObj = {
     code: 'ERR_INVALID_ARG_TYPE',
-    type: TypeError,
-    message: 'The first argument must be one of type string, Buffer, ' +
-             'ArrayBuffer, Array, or Array-like Object. Received ' +
-             `type ${actualType}`
-  });
-  throws(() => Buffer.from(input), err);
+    name: 'TypeError',
+    message: 'The first argument must be of type string or an instance of ' +
+             'Buffer, ArrayBuffer, or Array or an Array-like Object.' +
+             common.invalidArgTypeHelper(input)
+  };
+  throws(() => Buffer.from(input), errObj);
+  throws(() => Buffer.from(input, 'hex'), errObj);
 });
 
-[
-  new Number(true),
-  new MyBadPrimitive()
-].forEach((input) => {
-  const errMsg = common.expectsError({
-    code: 'ERR_INVALID_ARG_TYPE',
-    type: TypeError,
-    message: 'The "value" argument must not be of type number. ' +
-             'Received type number'
-  });
-  throws(() => Buffer.from(input), errMsg);
-});
+Buffer.allocUnsafe(10); // Should not throw.
+Buffer.from('deadbeaf', 'hex'); // Should not throw.
