@@ -20,7 +20,10 @@ assert(
 `;
 
 const args = ['--inspect', '-e', script];
-const child = spawn(process.execPath, args, { stdio: 'inherit' });
+const child = spawn(process.execPath, args, {
+  stdio: 'inherit',
+  env: { ...process.env, NODE_V8_COVERAGE: '' }
+});
 child.on('exit', (code, signal) => {
   process.exit(code || signal);
 });
